"""Objects used in the Apibara client."""

import base64
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Union


@dataclass(eq=True, frozen=True)
class EventFilter:
    """Filter transactions events based on their signature.
    
    The EventFilter can, optionally, filter by the emitting contract.
    
    Parameters
    ----------
    signature:
        the event signature.
    address:
        the contract address.
    """
    signature: str
    address: Optional[bytes]

    @classmethod
    def from_event_name(
        cls, name: str, address: Optional[Union[str, bytes, int]] = None
    ) -> "EventFilter":
        """Create an EventFilter from the event name and contract address."""
        if isinstance(address, str):
            address = bytes.fromhex(address.replace("0x", ""))
        if isinstance(address, int):
            address = address.to_bytes(32, "big")
        return cls(name, address)

    def to_json(self):
        """Returns the json representation of the filter."""
        return {"signature": self.signature, "address": self.address}

    @classmethod
    def from_json(cls, data):
        """Create an EventFilter from its json representation."""
        signature = data["signature"]
        if not isinstance(signature, str):
            raise ValueError("invalid signature. must be str")
        address = data.get("address")
        if address is None:
            return cls(signature, None)
        if not isinstance(address, bytes):
            raise ValueError("invalid address. must be bytes")
        return cls(signature, address)


@dataclass
class BlockHeader:
    """Information about a block.
    
    Parameters
    ----------
    hash:
        the block hash.
    parent_hash:
        the hash of the block's parent.
    number:
        the block number.
    timestamp:
        the block timestamp.
    """
    hash: bytes
    parent_hash: Optional[bytes]
    number: int
    timestamp: datetime

    @classmethod
    def from_proto(cls, block) -> "BlockHeader":
        hash = base64.b64decode(block["block_hash"]["hash"]).ljust(32, b"\0")
        parent_hash = base64.b64decode(block["parent_block_hash"]["hash"]).ljust(
            32, b"\0"
        )
        number = int(block.get("block_number", 0))
        timestamp = datetime.fromisoformat(block["timestamp"][:-1])
        return BlockHeader(
            hash=hash, parent_hash=parent_hash, number=number, timestamp=timestamp
        )

    def __str__(self) -> str:
        return f"BlockHeader(hash=0x{self.hash.hex()}, parent_hash=0x{self.parent_hash.hex()}, number={self.number}, timestamp={self.timestamp})"


@dataclass
class Event:
    """Base class for chain-specific events."""
    pass


@dataclass
class StarkNetEvent(Event):
    """StarkNet event.
    
    Parameters
    ----------
    name:
        event name.
    address:
        address of the contract emitting the event.
    log_index:
        index in the list of events emitted by the block.
    topics:
        event topics. Usually this is the hash of the event name.
    data:
        raw event data.
    transaction_hash:
        hash of the transaction emitting the event.
    """
    name: Optional[str]
    address: bytes
    log_index: int
    topics: List[bytes]
    data: List[bytes]
    transaction_hash: bytes

    @classmethod
    def from_proto(cls, event, log_index, event_name, tx_hash) -> "StarkNetEvent":
        address = base64.b64decode(event["from_address"]).ljust(32, b"\0")
        keys = [base64.b64decode(k).ljust(32, b"\0") for k in event["keys"]]
        data = [base64.b64decode(k).ljust(32, b"\0") for k in event["data"]]
        return StarkNetEvent(
            name=event_name,
            address=address,
            log_index=log_index,
            topics=keys,
            data=data,
            transaction_hash=tx_hash,
        )


@dataclass
class NewBlock:
    """Information about a new block being indexed.
    
    Parameters
    ----------
    new_head:
        block header.
    """
    new_head: BlockHeader


@dataclass
class Reorg:
    new_head: BlockHeader


@dataclass
class NewEvents:
    """Information about indexed events in a block.
    
    Parameters
    ----------
    block
        the block containing the events.
    events:
        list of events.
    """
    block: BlockHeader
    events: List[Event]



"""Apibara indexer runner."""

import asyncio
import base64
import warnings
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Awaitable, Callable, Generic, List, Optional, TypeVar

from grpc.aio import AioRpcError
from grpc_requests.aio import AsyncClient
from starknet_py.contract import ContractFunction

from apibara.indexer.storage import IndexerStorage, Storage
from apibara.logging import logger
from apibara.model import (BlockHeader, EventFilter, NewBlock, NewEvents,
                           Reorg, StarkNetEvent)

UserContext = TypeVar("UserContext")


@dataclass
class Info(Generic[UserContext]):
    """State shared between handlers.
    
    Parameters
    ----------
    context:
        application-specific context.
    storage:
        access the chain-aware storage.
    """
    context: UserContext
    storage: Storage

    _new_event_filters: List[EventFilter] = field(default_factory=list)

    def add_event_filters(self, filters: List[EventFilter]):
        """Add the provided event filters to the indexer.

        The indexer will re-scan the current block for any event
        matching the new filters.
        """
        self._new_event_filters.extend(filters)

    def _take_new_event_filters(self):
        """Returns and empties the current EventFilter list."""
        filters = self._new_event_filters
        self._new_event_filters = []
        return filters


NewEventsHandler = Callable[[Info, NewEvents], Awaitable[None]]
BlockHandler = Callable[[Info, NewBlock], Awaitable[None]]
ReorgHandler = Callable[[Info, Reorg], Awaitable[None]]


@dataclass
class IndexerRunnerConfiguration:
    """IndexerRunner configuration.
    
    Parameters
    ----------
    apibara_url:
        url of the Apibara stream.
    apibara_ssl:
        flag to connect using SSL.
    storage_url:
        MongoDB connection string, used to store the indexer  data and state.
    """
    apibara_url: Optional[str] = None
    apibara_ssl: bool = True
    rpc_url: Optional[str] = None
    storage_url: Optional[str] = None


class IndexerRunner(Generic[UserContext]):
    """Run an indexer, listening for new events and calling the provided callbacks.
    
    Parameters
    ----------
    indexer_id:
        unique id of this indexer. Used when persisting state.
    new_events_handler:
        async function called for each new block containing indexed events.
    reset_state:
        flag to restart the indexer from the beginning.
    config:
        options to set the input stream and connection string.
    """

    def __init__(
        self,
        *,
        indexer_id: str,
        new_events_handler: NewEventsHandler,
        reset_state: bool = False,
        config: Optional[IndexerRunnerConfiguration] = None,
    ) -> None:
        if config is None:
            config = IndexerRunnerConfiguration()
        self._reset_state = reset_state
        self._indexer_id = indexer_id
        self._config = config
        self._new_events_handler = new_events_handler
        self._block_handler = None
        self._reorg_handler = None
        self._context = None
        self._indexer_storage = IndexerStorage(
            self._config.storage_url, self._indexer_id
        )

        self._indexer_config = None
        self._event_name_map = dict()

    def create_if_not_exists(
        self,
        filters: Optional[List[EventFilter]] = None,
        index_from_block: Optional[int] = None,
    ):
        self.add_event_filters(filters, index_from_block)

    def add_event_filters(
        self,
        filters: Optional[List[EventFilter]] = None,
        index_from_block: Optional[int] = None,
    ):
        """Add the initial event filters.
        
        Parameters
        ----------
        filters:
            a list of EventFilter
        index_from_block:
            start indexing from the given block.
        """
        self._indexer_config = {
            "filters": filters,
            "index_from_block": index_from_block,
        }

    def set_context(self, context: UserContext):
        """Set the context object used to share information between handlers."""
        self._context = context

    def add_block_handler(self, block_handler: BlockHandler) -> None:
        """Add a callback called every time there is a new block."""
        self._block_handler = block_handler

    def add_reorg_handler(self, reorg_handler: ReorgHandler) -> None:
        """Add a callback called every time there is a chain reorganization."""
        self._reorg_handler = reorg_handler

    async def run(self):
        """Run the indexer until stopped."""
        self._print_deprecation_notice()
        self._check_config()
        if self._reset_state:
            self._delete_old_state()
        self._write_initial_state()
        await self._do_run()

    def _check_config(self):
        if self._config is None:
            raise RuntimeError("must provide config to IndexerRunner")

        if self._config.apibara_url is None:
            raise RuntimeError("must provide an apibara_url in config")

        if self._config.storage_url is None:
            raise RuntimeError("must provide a storage_url in config")

    def _print_deprecation_notice(self):
        warnings.warn(
            "IndexerRunner is deprecated and will be removed in the next release.",
            DeprecationWarning,
        )

    def _delete_old_state(self):
        self._indexer_storage.drop_database()

    def _write_initial_state(self):
        self._indexer_storage.initialize(
            self._indexer_config["index_from_block"], self._indexer_config["filters"]
        )

    async def _do_run(self):
        client = AsyncClient(self._config.apibara_url, ssl=self._config.apibara_ssl)

        node_service = await client.service("apibara.node.v1alpha1.Node")
        starting_sequence = self._indexer_storage.starting_sequence()
        filters_def = self._indexer_storage.event_filters()
        filters = [CompiledEventFilter.from_event_filter(f) for f in filters_def]

        message_stream = await node_service.StreamMessages(
            {"starting_sequence": starting_sequence}
        )

        while True:
            # We expect one heartbeat every 30 seconds
            # Add 15 seconds of buffer, we the stream doesn't produce any message
            # in this timeframe then there is something wrong with the stream.
            message = await asyncio.wait_for(message_stream.__anext__(), timeout=45.0)

            if "data" in message:
                block = message["data"]["data"]
                block_header = BlockHeader.from_proto(block)
                with self._block_context(block_header.number) as info:
                    # new filters added across all loops of this block
                    new_filters = []
                    # filters for this specific loop
                    loop_filters = filters
                    while loop_filters:
                        events = self._block_events_matching_filters(
                            loop_filters, block
                        )
                        new_events = NewEvents(block=block_header, events=events)

                        if self._block_handler:
                            new_block = NewBlock(new_head=block_header)
                            await self._block_handler(info, new_block)
                        if new_events.events:
                            await self._new_events_handler(info, new_events)

                        if info._new_event_filters:
                            new_loop_filters = info._take_new_event_filters()
                            new_filters.extend(new_loop_filters)

                            loop_filters = [
                                CompiledEventFilter.from_event_filter(f)
                                for f in new_loop_filters
                            ]
                        else:
                            loop_filters = False

                    # finished parsing events. update state with the new filters
                    if new_filters:
                        filters_def.extend(new_filters)
                        filters_def = _unique_event_filters(filters_def)
                        self._indexer_storage._set_event_filters(
                            filters_def, session=info.storage._session
                        )
                        filters = [
                            CompiledEventFilter.from_event_filter(f)
                            for f in filters_def
                        ]

            elif "invalidate" in message:
                raise RuntimeError("reorg are not expected on StarkNet")

    @contextmanager
    def _block_context(self, number: int) -> Info[UserContext]:
        with self._indexer_storage.create_storage_for_block(number) as storage:
            yield Info(context=self._context, storage=storage)

    def _block_events_matching_filters(self, filters, block):
        matched_events = []
        log_index = 0
        transactions = block.get("transactions", [])
        for receipt in block.get("transaction_receipts",[]):
            tx = transactions[int(receipt.get("transaction_index", 0))]
            if "events" in receipt:
                for event in receipt["events"]:
                    event_name, matched = self._filter_matching(filters, event)
                    if matched:
                        tx_hash = _transaction_hash(tx)
                        event = StarkNetEvent.from_proto(
                            event, log_index, event_name, tx_hash
                        )
                        matched_events.append(event)
                    log_index += 1
        return matched_events

    def _filter_matching(self, filters, event):
        for filter in filters:
            if filter.matches(event):
                return filter.name, True
        return None, False


@dataclass
class CompiledEventFilter:
    name: str
    keys: List[bytes]
    address: Optional[bytes]

    @classmethod
    def from_event_filter(cls, filter: EventFilter):
        address = None
        if filter.address is not None:
            address = filter.address.ljust(32, b"\0")

        return CompiledEventFilter(
            name=filter.signature,
            keys=[ContractFunction.get_selector(filter.signature).to_bytes(32, "big")],
            address=address,
        )

    def matches(self, event):
        if "keys" not in event:
            return False

        if self.address is not None and "from_address" in event:
            event_address = base64.b64decode(event["from_address"]).ljust(32, b"\0")
            if event_address != self.address:
                return False

        event_keys = [base64.b64decode(k).ljust(32, b"\0") for k in event["keys"]]

        return event_keys == self.keys


def _transaction_hash(tx) -> bytes:
    common = None
    if "invoke" in tx:
        common = tx["invoke"]["common"]
    elif "deploy" in tx:
        common = tx["deploy"]["common"]
    elif "declare" in tx:
        common = tx["declare"]["common"]
    elif "l1_handler" in tx:
        common = tx["l1_handler"]["common"]
    else:
        raise RuntimeError("unknown transaction type")
    return base64.b64decode(common["hash"])


def _unique_event_filters(filters: List[EventFilter]) -> List[EventFilter]:
    return list(set(filters))



from contextlib import contextmanager
from typing import Any, Iterable, Iterator, List, Optional

from pymongo import MongoClient
from pymongo.client_session import ClientSession
from pymongo.database import Database

from apibara.model import EventFilter

Document = dict[str, Any]
Filter = dict[str, Any]
Update = dict[str, Any]
Projection = dict[str, any]


class IndexerStorage:
    """Create instances of Storage."""

    def __init__(self, url: Optional[str], indexer_id: str) -> None:
        if url is None:
            raise ValueError("Storage url must be not None")

        self.db_name = indexer_id.replace("-", "_")
        self._indexer_id = indexer_id

        self._mongo = MongoClient(url)
        self.db = self._mongo[self.db_name]

    @contextmanager
    def create_storage_for_block(self, block_number: int) -> Iterator["Storage"]:
        with self._mongo.start_session() as session:
            yield Storage(self.db, session, block_number)
            self._update_indexed_to(block_number, session)

    def initialize(self, starting_sequence: int, filters: List[EventFilter]):
        existing = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
        if existing is not None:
            return
        self.db["_apibara"].insert_one(
            {
                "indexer_id": self._indexer_id,
                "indexed_to": starting_sequence,
                "filters": [f.to_json() for f in filters],
            }
        )

    def starting_sequence(self):
        state = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
        if state is None or state.get("indexed_to") is None:
            return None
        return state["indexed_to"] + 1

    def event_filters(self):
        state = self.db["_apibara"].find_one({"indexer_id": self._indexer_id})
        if state is None:
            raise RuntimeError("indexer state not found")
        return [EventFilter.from_json(j) for j in state["filters"]]

    def _set_event_filters(self, filters: List[EventFilter], session: ClientSession):
        """Set the indexer event filters, overriding the previous filters."""
        filters_docs = [f.to_json() for f in filters]
        self.db["_apibara"].update_one(
            {"indexer_id": self._indexer_id},
            {"$set": {"filters": filters_docs}},
            session=session,
        )

    def drop_database(self):
        self._mongo.drop_database(self.db_name)

    def _update_indexed_to(self, block_number: int, session: ClientSession):
        self.db["_apibara"].update_one(
            {"indexer_id": self._indexer_id},
            {"$set": {"indexed_to": block_number}},
            upsert=True,
            session=session,
        )


class Storage:
    """Chain-aware document storage."""

    def __init__(self, db: Database, session: ClientSession, block_number: int) -> None:
        self._db = db
        self._block_number = block_number
        self._session = session

    async def insert_one(self, collection: str, doc: Document):
        """Insert `doc` into `collection`."""
        self._add_chain_information(doc)
        self._db[collection].insert_one(doc, session=self._session)

    async def insert_many(self, collection: str, docs: Iterable[Document]):
        """Insert multiple `docs` into `collection`."""
        for doc in docs:
            self._add_chain_information(doc)
        self._db[collection].insert_many(docs, session=self._session)

    async def delete_one(self, collection: str, filter: Filter):
        """ "Delete the first document in `collection` matching `filter`."""
        self._add_current_block_to_filter(filter)
        self._db[collection].update_one(
            filter,
            {"$set": {"_chain.valid_to": self._block_number}},
            session=self._session,
        )

    async def delete_many(self, collection: str, filter: Filter):
        """Delete all documents in `collection` matching `filter`."""
        self._add_current_block_to_filter(filter)
        self._db[collection].update_many(
            filter,
            {"$set": {"_chain.valid_to": self._block_number}},
            session=self._session,
        )

    async def find_one(self, collection: str, filter: Filter) -> Optional[Document]:
        """Find the first document in `collection` matching `filter`."""
        self._add_current_block_to_filter(filter)
        return self._db[collection].find_one(filter, session=self._session)

    async def find(
        self,
        collection: str,
        filter: Filter,
        sort: Optional[dict[str, int]] = None,
        projection: Optional[Projection] = None,
        skip: int = 0,
        limit: int = 0,
    ) -> Iterable[dict]:
        """Find all documents in `collection` matching `filter`.

        Arguments
        ---------
        - `collection`: the collection,
        - `filter`: the filter,
        - `sort`: keys used for sorting, e.g. `{"a": -1}` sorts documents by key `a` in descending order,
        - `project`: filter document keys to reduce the document size,
        - `skip`: number of documents to skip,
        - `limit`: maximum number of documents returned.
        """
        self._add_current_block_to_filter(filter)
        cursor = self._db[collection].find(
            filter, projection, skip, limit, session=self._session
        )
        if sort is not None:
            for field, order in sort.items():
                cursor = cursor.sort(field, order)
        return cursor

    async def find_one_and_replace(
        self,
        collection: str,
        filter: Filter,
        replacement: Document,
        upsert: bool = False,
    ):
        """Replace the first document in `collection` matching `filter` with `replacement`.

        If `upsert = True`, insert `replacement` even if no document matched the `filter`.
        """
        # Step 1. Update the old document (if any) by clamping its validity range
        self._add_current_block_to_filter(filter)
        existing = self._db[collection].find_one_and_update(
            filter,
            {"$set": {"_chain.valid_to": self._block_number}},
            session=self._session,
        )

        # Step 2. Insert the new document.
        # Insert only if the existing document exists or if upsert.
        if existing is not None or upsert:
            await self.insert_one(collection, replacement)

        return existing

    async def find_one_and_update(
        self, collection: str, filter: Filter, update: Update
    ):
        """Update the first document in `collection` matching `filter` with `update`."""
        # Step 1. Update the old document (if any) by clamping its validity range
        self._add_current_block_to_filter(filter)
        existing = self._db[collection].find_one_and_update(
            filter,
            {"$set": {"_chain.valid_to": self._block_number}},
            session=self._session,
        )

        # Step 2. To simulate an update, first insert then call update on it.
        if existing is not None:
            del existing["_id"]
            del existing["_chain"]
            await self.insert_one(collection, existing)
            self._db[collection].update_one(filter, update, session=self._session)

        return existing

    def _add_chain_information(self, doc: Document):
        doc["_chain"] = {"valid_from": self._block_number, "valid_to": None}

    def _add_current_block_to_filter(self, filter: Filter):
        filter["_chain.valid_to"] = None
